import 'dart:math';

import 'package:prefix_sum/prefix_sum.dart';
import 'package:test/test.dart';

void main() {
  group('SparseBinaryIndexTree', () {
    test('should act like an ordinary List<num>', () {
      expect(SparsePrefixSumList(1) is List<num>, true);

      final random = Random();
      final length = random.nextInt(10) + 1;
      final defaultValue = random.nextInt(100);
      final list = SparsePrefixSumList(defaultValue, length);

      for (int i = 0; i < length; i++) {
        expect(list[i], defaultValue);
      }
      for (int i = 0; i < 10; i++) {
        final newValue = random.nextInt(100);
        final newIndex = random.nextInt(length);
        list[newIndex] = newValue;
        expect(list[newIndex], newValue);
      }
    });

    test('should contain a prefix sum index', () {
      final random = Random();
      for (int i = 0; i < 20; i++) {
        final defaultValue = random.nextInt(100) + 1;
        final bit = SparsePrefixSumList(defaultValue, random.nextInt(100) + 1);

        for (int i = 0; i < random.nextInt(10) + 1; i++) {
          final index = random.nextInt(bit.length);
          expect(bit.getPrefixSum(index), (index + 1) * defaultValue);
          expect(bit.findPrefixSumBefore(index * defaultValue), index - 1);
        }

        for (int i = 0; i < random.nextInt(10) + 1; i++) {
          final index = random.nextInt(bit.length);
          expect(bit.getPrefixSum(index), (index + 1) * defaultValue);
          expect(bit.findPrefixSumBefore(index * defaultValue), index - 1);
        }

        final mirror =
            List<num>.filled(bit.length, defaultValue, growable: true);
        for (int i = 0; i < random.nextInt(bit.length) + 1; i++) {
          final value = random.nextInt(100) + 1;
          mirror.add(value);
          bit.add(value);
        }
        final mirrorSum = sumList(mirror);

        for (int i = 0; i < random.nextInt(10) + 1; i++) {
          final index = random.nextInt(bit.length);
          expect(bit.getPrefixSum(index), mirrorSum[index]);
        }

        for (int i = 0; i < random.nextInt(10) + 1; i++) {
          final expectedIndex = random.nextInt(bit.length);
          final sumToFind = mirrorSum[expectedIndex] -
              mirror[expectedIndex] * random.nextDouble();
          expect(bit.findPrefixSumBefore(sumToFind), expectedIndex - 1);
        }
      }
    });
  });
}

List<num> sumList(List<num> from) {
  final sumList = <num>[from[0]];
  from.reduce((a, b) {
    final next = a + b;
    sumList.add(next);
    return next;
  });
  return sumList;
}
